#include "comm.h"

VOID
Fatal( LPCTSTR szErrorMessage, LPCTSTR szCaption )
{
	static char buf[128];
	wsprintf( buf, "%s (%d)", szErrorMessage, GetLastError() );
	MessageBox( HWND_DESKTOP, buf, szCaption, MB_OK );
	ExitProcess(1);
}

LPCOMMINFO
InitCommInfo( HWND hWnd )
{
	LPCOMMINFO lpCom = 
		(LPCOMMINFO) LocalAlloc( LPTR, sizeof( COMMINFO ) );

	if ( lpCom == NULL )
	{
		Fatal( "LocalAlloc failed", "InitCommInfo" );
	}
			
	// initialize comm info structure

	lpCom->idComDev			= 0 ;
	lpCom->bPort			= 1 ;
	lpCom->fConnected		= FALSE ;
	lpCom->fXonXoff			= FALSE ;
	lpCom->fUseCNReceive		= TRUE ;
	lpCom->fDisplayErrors		= TRUE ;
	lpCom->bByteSize 		= 8 ;
	lpCom->bFlowCtrl		= FC_RTSCTS ;
	lpCom->bParity			= NOPARITY ;
	lpCom->bStopBits		= ONESTOPBIT ;
	lpCom->dwBaudRate 		= CBR_9600 ;
	lpCom->osWrite.Offset		= 0 ;
	lpCom->osWrite.OffsetHigh	= 0 ;
	lpCom->osRead.Offset		= 0 ;
	lpCom->osRead.OffsetHigh	= 0 ;

	// create I/O events used for overlapped reads / writes

	lpCom->osRead.hEvent = CreateEvent( NULL, // no security
                                              TRUE,    // explicit reset req.
                                              FALSE,   // initial event reset
                                              NULL ) ; // no name

	if ( lpCom->osRead.hEvent == NULL )
	{
		Fatal( "CreateEvent (osRead) failed", "InitCommInfo" );
	}

	lpCom->osWrite.hEvent = CreateEvent( NULL, // no security
                                              TRUE,     // explicit reset req.
                                              FALSE,    // initial event reset
                                              NULL ) ;  // no name

	if ( lpCom->osWrite.hEvent == NULL )
	{
		Fatal( "CreateEvent (osWrite) failed", "InitCommInfo" );
	}

	lpCom->hPostEvent =		NULL ;
	lpCom->hWatchThread =		NULL ;
	lpCom->hWatchEvent =		NULL ;
	lpCom->dwThreadID =		0 ;
	lpCom->hWnd =			hWnd ;
}

VOID
DestroyCommInfo( LPCOMMINFO lpCom )
{
	// force connection closed (if not already closed)

	if ( lpCom->fConnected )
	{
      		CloseConnection( lpCom ) ;
	}

	// clean up event objects

	CloseHandle( lpCom->osRead.hEvent ) ;
	CloseHandle( lpCom->osWrite.hEvent ) ;
	CloseHandle( lpCom->hPostEvent ) ;

	LocalFree( lpCom ) ;
}

BOOL
OpenConnection( LPCOMMINFO lpCom )
{
	const char	COMPREFIX [] = "COM" ;

	char		szPort[ 15 ] ;
	BOOL		fRetVal ;
	HANDLE 		hCommWatchThread ;
	DWORD		dwThreadID ;
	COMMTIMEOUTS	CommTimeOuts ;

	// append port number to the COM prefix

	wsprintf( szPort, "%s%d", COMPREFIX, lpCom->bPort ) ;

	// open COMM device

	lpCom->idComDev = 
		CreateFile( szPort,
			GENERIC_READ | GENERIC_WRITE,
			0,			// exclusive access
			NULL,			// no security attrs
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL |
			FILE_FLAG_OVERLAPPED,	// overlapped I/O
			NULL );

	if ( lpCom->idComDev == INVALID_HANDLE_VALUE )
	{
		return ( FALSE ) ;
	}
	else
	{
		// get any early notifications

		SetCommMask( lpCom->idComDev, EV_RXCHAR ) ;

		// setup device buffers

		SetupComm( lpCom->idComDev, 4096, 4096 ) ;

		// purge any information in the buffer

		PurgeComm( lpCom->idComDev,
			PURGE_TXABORT | PURGE_RXABORT |
			PURGE_TXCLEAR | PURGE_RXCLEAR ) ;

		// set up for overlapped I/O

		CommTimeOuts.ReadIntervalTimeout = 0xFFFFFFFF ;
		CommTimeOuts.ReadTotalTimeoutMultiplier = 0 ;
		CommTimeOuts.ReadTotalTimeoutConstant = 1000 ;

		// CBR_9600 is approximately 1byte/ms. For our purposes,
		// allow double the expected time per character for a
		// fudge factor.

		CommTimeOuts.WriteTotalTimeoutMultiplier = 
			2 * CBR_9600 / lpCom->dwBaudRate ;

		CommTimeOuts.WriteTotalTimeoutConstant = 0 ;

		SetCommTimeouts( lpCom->idComDev, &CommTimeOuts ) ;
	}

	fRetVal = SetupConnection( lpCom ) ;

	if ( fRetVal )
	{
		lpCom->fConnected = TRUE ;

		// Create a secondary thread to watch for an event.

		hCommWatchThread =
			CreateThread(
				(LPSECURITY_ATTRIBUTES) NULL,
				0,
				(LPTHREAD_START_ROUTINE) CommWatchProc,
				(LPVOID) lpCom,
				0,
				&dwThreadID
			) ;

		if ( hCommWatchThread == NULL )
		{
			lpCom->fConnected = FALSE ;
			CloseHandle( lpCom->idComDev ) ;
			fRetVal = FALSE ;
		}
		else
		{
			lpCom->dwThreadID = dwThreadID ;
			lpCom->hWatchThread = hCommWatchThread ;

			// assert DTR

			EscapeCommFunction( lpCom->idComDev, SETDTR ) ;

		}
	}
	else
	{
		lpCom->fConnected = FALSE ;
		CloseHandle( lpCom->idComDev ) ;
	}

	return ( fRetVal ) ;
}

BOOL 
SetupConnection( LPCOMMINFO lpCom )
{
	BOOL	fRetVal ;
	BYTE	bSet ;
	DCB	dcb ;

	dcb.DCBlength = sizeof( DCB ) ;

	GetCommState( lpCom->idComDev, &dcb ) ;

	dcb.BaudRate	= lpCom->dwBaudRate ;
	dcb.ByteSize	= lpCom->bByteSize ;
	dcb.Parity 	= lpCom->bParity ;
	dcb.StopBits	= lpCom->bStopBits ;

	// setup hardware flow control

	bSet = (BYTE) (( lpCom->bFlowCtrl & FC_DTRDSR ) != 0) ;
	dcb.fOutxDsrFlow = bSet ;
	if (bSet)
	{
		dcb.fDtrControl = DTR_CONTROL_HANDSHAKE ;
	}
	else
	{
		dcb.fDtrControl = DTR_CONTROL_ENABLE ;
	}

	bSet = (BYTE) (( lpCom->bFlowCtrl & FC_RTSCTS) != 0) ;
	dcb.fOutxCtsFlow = bSet ;
	if (bSet)
	{
		dcb.fRtsControl = RTS_CONTROL_HANDSHAKE ;
	}
	else
	{
		dcb.fRtsControl = RTS_CONTROL_ENABLE ;
	}

	// setup software flow control

	bSet = (BYTE) (( lpCom->bFlowCtrl & FC_XONXOFF) != 0) ;

	dcb.fInX = dcb.fOutX = bSet ;
	dcb.XonChar = ASCII_XON ;
	dcb.XoffChar = ASCII_XOFF ;
	dcb.XonLim = 100 ;
	dcb.XoffLim = 100 ;

	// other various settings

	dcb.fBinary = TRUE ;
	dcb.fParity = TRUE ;

	fRetVal = SetCommState( lpCom->idComDev, &dcb ) ;

	return ( fRetVal ) ;
}

BOOL 
CloseConnection( LPCOMMINFO lpCom )
{
	// set connected flag to FALSE

	lpCom->fConnected = FALSE ;

	// disable event notification and wait for thread to halt

	SetCommMask( lpCom->idComDev, 0 ) ;

	// block until thread has been halted

	while ( lpCom->dwThreadID != 0 )
	{
		;
	}

	// drop DTR

	EscapeCommFunction( lpCom->idComDev, CLRDTR ) ;

	// purge any outstanding reads/writes and close device handle

	PurgeComm( lpCom->idComDev, PURGE_TXABORT | PURGE_RXABORT |
			PURGE_TXCLEAR | PURGE_RXCLEAR ) ;

	CloseHandle( lpCom->idComDev ) ;

	return ( TRUE ) ;
}

VOID
DisplayCommError( DWORD dwError, LPCOMMINFO lpCom )
{
	char	szError[ 10 ] ;

	if (   ( lpCom == NULL ) ||
	     ( ( lpCom != NULL ) && lpCom->fDisplayErrors &&
	       ( dwError > 0 ) ) )
	{
		wsprintf( szError, "<CE-%u>", dwError ) ;
		MessageBox( HWND_DESKTOP, szError, "CommError", MB_OK ) ;
	}
}

INT
HandleReadCommError( LPCOMMINFO lpCom, LPCOMSTAT lpComStat )
{
	DWORD	dwLength ;
	DWORD	dwErrorFlags ;
	DWORD	dwError ;

	if ( GetLastError() == ERROR_IO_PENDING )
	{
		// We have to wait for read to complete.
		// This function will timeout according to the
		// CommTimeOuts.ReadTotalTimeoutConstant variable
		// Every time it times out, check for port errors

		while ( ! GetOverlappedResult( lpCom->idComDev,
			   &lpCom->osRead, &dwLength, TRUE ) )
		{
			dwError = GetLastError() ;
			if ( dwError == ERROR_IO_INCOMPLETE )
			{
				// normal result if not finished
				continue;
			}
			else
			{
				// an error occurred, try to recover

				DisplayCommError( dwError, NULL ) ;

				ClearCommError( lpCom->idComDev,
					&dwErrorFlags, lpComStat ) ;

				DisplayCommError( dwErrorFlags, lpCom ) ;
				break;
			}
		}
	}
	else
	{
		// some other error occurred
		dwLength = 0 ;
		ClearCommError( lpCom->idComDev, &dwErrorFlags, lpComStat ) ;
		DisplayCommError( dwErrorFlags, lpCom ) ;
	}

	return dwLength;
}

INT
ReadCommBlock( LPCOMMINFO lpCom, LPSTR lpszBlock, INT nMaxLength )
{
	BOOL		fReadStat ;
	COMSTAT		ComStat ;
	DWORD		dwErrorFlags ;
	DWORD		dwLength ;

	// only try to read number of bytes in queue

	ClearCommError( lpCom->idComDev, &dwErrorFlags, &ComStat ) ;
	dwLength = min( (DWORD) nMaxLength, ComStat.cbInQue ) ;

	if ( dwLength > 0 )
	{
		fReadStat = ReadFile( lpCom->idComDev, lpszBlock,
				dwLength, &dwLength, &lpCom->osRead ) ;
		if ( ! fReadStat )
		{
			dwLength = HandleCommError( lpCom, &ComStat );
		}
	}

	return ( dwLength ) ;
}

BOOL
HandleWriteCommError( LPCOMMINFO lpCom, LPCOMSTAT lpComStat )
{
	DWORD		dwBytesWritten ;
	DWORD		dwErrorFlags ;
	DWORD		dwError ;
	DWORD		dwBytesSent = 0 ;
	char		szError[ 128 ] ;

	if ( GetLastError() == ERROR_IO_PENDING )
	{
		// We should wait for the completion of the write operation
		// so we know if it worked or not

		// This is only one way to do this. It might be beneficial to
		// place the write operation in a separate thread
		// so that blocking on completion will not negatively
		// affect the responsiveness of the UI

		// If the write takes too long to complete, this
		// function will timeout according to the
		// CommTimeOuts.WriteTotalTimeoutMultiplier variable.
		// This code logs the timeout but does not retry
		// the write.

		while ( ! GetOverlappedResult( lpCom->idComDev,
			   &lpCom->osWrite, &dwBytesWritten, TRUE ) )
		{
			dwError = GetLastError();
			if ( dwError == ERROR_IO_INCOMPLETE )
			{
				// normal result if not finished
				dwBytesSent += dwBytesWritten;
				continue;
			}
			else
			{
				// an error occurred, try to recover

				DisplayCommError( dwError, NULL ) ;

				ClearCommError( lpCom->idComDev,
					&dwErrorFlags, lpComStat ) ;

				DisplayCommError( dwError, lpCom ) ;
				break;
			}
		}
		dwBytesSent += dwBytesWritten;
	}
	else
	{
		// some other error occurred
		ClearCommError( lpCom->idComDev, &dwErrorFlags, lpComStat ) ;

		DisplayCommError( dwErrorFlags, lpCom );

		return ( FALSE );
	}

	return ( TRUE );
}

BOOL
WriteCommBlock( LPCOMMINFO lpCom, LPSTR lpByte, DWORD dwBytesToWrite )
{
	BOOL		bStatus = TRUE ;
	BOOL		fWriteStat ;
	DWORD		dwBytesWritten ;
	COMSTAT		ComStat ;

	fWriteStat = WriteFile( lpCom->idComDev, lpByte, dwBytesToWrite,
			&dwBytesWritten, &lpCom->osWrite );

	// Note that normally the code will not execute the following
	// because the driver caches write operations. Small I/O requests
	// (up to several thousand bytes) will normally be accepted
	// immediately and WriteFile will return true even though an
	// overlapped operation was specified

	if ( ! fWriteStat )
	{
		bStatus = HandleCommWriteError( lpCom, &ComStat );
	}

	return ( bStatus ) ;
}

VOID
CommWatchNotify( HWND hWnd, LPCOMMDATA lpData )
{
	LPCOMMDATA lpCopy ;

	lpCopy = (LPCOMMDATA) LocalAlloc( LPTR, sizeof( COMMDATA ) );
	if ( lpCopy == NULL )
	{
		Fatal( "LocalAlloc failed", "CommWatchCopyData" );
	}

	memcpy( lpCopy, lpData, sizeof( COMMDATA ) );

	SendMessage( hWnd, MNO_COMMDATA, (WPARAM) 0, (LPARAM) lpCopy ) ;
}
	
VOID
CommWatchRead( LPCOMMINFO lpCom )
{
	COMMDATA comRecvd ;

	do
	{
		if ( ( comRecvd.nLength = ReadCommBlock( lpCom,
			(LPSTR) comRecvd.abIn, MAXBLOCK ) ) > 0 )
		{
			CommWatchNotify( lpCom->hWnd, &comRecvd );
		}
	}
	while ( comRecvd.nLength > 0 ) ;
}

DWORD WINAPI
CommWatchProc( LPVOID lpData )
{
	LPCOMMINFO lpCom;
	DWORD dwEvtMask ;
	OVERLAPPED os ;

	lpCom = (LPCOMMINFO) lpData;

	memset( &os, 0, sizeof( OVERLAPPED ) ) ;

	// create I/O event used for overlapped read

	os.hEvent = CreateEvent( NULL,    // no security
				 TRUE,    // explicit reset req
				 FALSE,   // initial event reset
				 NULL ) ; // no name
	if (os.hEvent == NULL)
	{
		Fatal( "Failed to create thread event", "CommWatchProc" ) ;
	}

	if ( ! SetCommMask( lpCom->idComDev, EV_RXCHAR ) )
	{
		return ( FALSE ) ;
	}

	while ( lpCom->fConnected )
	{
		dwEvtMask = 0 ;

		WaitCommEvent( lpCom->idComDev, &dwEvtMask, NULL );

		if ( ( dwEvtMask & EV_RXCHAR ) == EV_RXCHAR )
		{
			CommWatchRead( lpCom ) ;
		}
	}

	// get rid of event handle

	CloseHandle( os.hEvent ) ;

	// clear information in structure (kind of a "we're done flag")

	lpCom->dwThreadID =		0 ;
	lpCom->hWatchThread =		NULL ;

	return ( TRUE ) ;
}

